#!/usr/bin/env python
#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
# 
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

import os
import sqlite3
import subprocess
import sys

from Bio import SeqIO

from graph import Graph

def get_frequencies(db):
	sql = """
		SELECT edge, COUNT(*)
		FROM states JOIN samples ON states.id=samples.id
		WHERE samples.accept=1
		GROUP BY edge;"""
	return dict(db.execute(sql))


def posterior_plot(frequencies, graph, prefix):
	norm = 1.0 / max(frequencies.itervalues())
	for e in frequencies:
		frequencies[e] *= norm
	graph.write_dot(prefix+'-posteriors.dot', frequencies)
	subprocess.check_call(
		['dot', '-Tsvg', '-Kneato', prefix+'-posteriors.dot'],
		stdout=open(prefix+'-posteriors.svg', 'w'))


def majority_rule_plot(frequencies, graph, threshold, nsamples, prefix):
	colors = {}
	posteriors = {}
	norm = 1.0 / nsamples
	threshold *= nsamples
	for e, f in frequencies.iteritems():
		if f > threshold:
			colors[e] = 'black'
			posteriors[e] = f * norm
		else:
			colors[e] = 'grey90'
	graph.write_dot(prefix+'-majority.dot', colors=colors)
	subprocess.check_call(
		['dot', '-Tsvg', '-Kneato', prefix+'-majority.dot'],
		stdout=open(prefix+'-majority.svg', 'w'))
	with open(prefix+'-majority.fa', 'w') as f:
		for record in graph.contigs_annotated(posteriors):
			SeqIO.write(record, f, 'fasta')


def consensus_plots(graph, db, threshold, prefix):
	sql = """
		SELECT COUNT(*)
		FROM samples
		WHERE accept=1;"""
	nsamples = db.execute(sql).fetchone()[0]
	frequencies = get_frequencies(db)
	posterior_plot(frequencies.copy(), graph, prefix)
	majority_rule_plot(frequencies, graph, threshold, nsamples, prefix)
	return frequencies


if __name__ == '__main__':
	graph = Graph(sys.argv[1])
	db = sqlite3.connect(sys.argv[2])
	threshold = float(sys.argv[3])
	consensus_plots(graph, db, threshold, 'consensus')

