#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
# 
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

import os
import sqlite3

from biolite import utils

from sampler import Sample
 
_db = None
_inserts = 0

samples_schema = (
	('id', 'INTEGER PRIMARY KEY NOT NULL'),
	('posterior', 'REAL'),
	('likelihood', 'REAL'),
	('prior', 'REAL'),
	('length', 'INTEGER'),
	('ncontigs', 'INTEGER'),
	('perturb', 'INTEGER'),
	('accept', 'INTEGER'),
	('runtime', 'REAL'))

states_schema = (
	('id', 'INTEGER'),
	('edge', 'INTEGER'))


def _create_sql(name, schema, index):
	sql = [" CREATE TABLE %s (" % name]
	sql += ["   %s %s," % s for s in schema[:-1]]
	sql += ["   %s %s);" % schema[-1]]
	sql += [" CREATE INDEX {0}_{1} ON {0}({1});".format(name, i) for i in index]
	return '\n'.join(sql)

def open(path):
	"""
	Establish a gobal database connection and create the tables if the
	database did not exist.
	"""
	global _db
	exists = os.path.exists(path)
	_db = sqlite3.connect(path, isolation_level=None)
	if not exists:
		_db.executescript(_create_sql(
			'samples', samples_schema, ['posterior', 'accept', 'runtime']))
		_db.executescript(_create_sql(
			'states', states_schema, ['id', 'edge']))
		return None
	else:
		last = int(next(_db.execute("SELECT MAX(id) FROM samples;"))[0])
		utils.info(
			"restarting from previous trace '%s' at iteration %d" % (
			path, last))
		return last

def last_accepted():
	return int(next(_db.execute("SELECT MAX(id) FROM samples WHERE accept=1;"))[0])

def get_sample(id):
    return Sample(*next(_db.execute("SELECT likelihood, prior, length, ncontigs FROM samples WHERE id=?", (id,))))

def get_state(state, id):
	for row in _db.execute("SELECT edge FROM states WHERE states.id=?", (id,)):
		state[row[0]-1] = True

def load(path):
	"""
	Establish a global database connection to an existing trace.
	"""
	global _db
	_db = sqlite3.connect(path, isolation_level=None)

def add(id, sample, perturb, state, accept, runtime):
	global _db, _inserts
	# Insert with transaction.
	if _inserts % 100 == 0:
		_db.execute("BEGIN")
	_db.execute(
		"INSERT INTO samples VALUES (?,?,?,?,?,?,?,?,?);",
		(id, sample.posterior, sample.likelihood, sample.prior, sample.length, sample.ncontigs, perturb, accept, runtime))
	for e, active in enumerate(state, start=1):
		if active:
			_db.execute("INSERT INTO states VALUES(?,?);", (id, e))
	if _inserts % 100 == 99:
		_db.execute("COMMIT")
	_inserts += 1

